@@ -2,10 +2,11 @@ use super::Card;
 use crate::VarName;
 use std::str::FromStr;
 
+/// Cao-lang functions
 #[derive(Debug, Clone, Default)]
 #[cfg_attr(feature = "serde", derive(serde::Serialize, serde::Deserialize))]
 pub struct Lane {
-    pub name: Option<String>,
+    pub name: String,
     #[cfg_attr(feature = "serde", serde(default = "Vec::new"))]
     pub arguments: Vec<VarName>,
     #[cfg_attr(feature = "serde", serde(default = "Vec::new"))]
@@ -13,9 +14,17 @@ pub struct Lane {
 }
 
 impl Lane {
+    #[must_use]
+    pub fn from_name<S: Into<String>>(name: S) -> Self {
+        Self {
+            name: name.into(),
+            ..Default::default()
+        }
+    }
+
     #[must_use]
     pub fn with_name<S: Into<String>>(mut self, name: S) -> Self {
-        self.name = Some(name.into());
+        self.name = name.into();
         self
     }
 
